local debugmode = false -- also shows friendly units

local addon = PvPTimer

local L = addon.Locale
local fb = addon.Lib.Bars
local media = addon.Lib.Media
local ACR = addon.Lib.AceConfigRegistry
local ACD = addon.Lib.AceConfigDialog
local LibStub = LibStub

-- global tables
addon.Anchors = {}
addon.AnchorIcons = {}
addon.Interface = {}
addon.Units = {}
addon.Specs = {}
addon.AnchorsLocked = true
-- addon.CopyHeaderSettings = nil
-- addon.CopyBarSettings = nil

-- copy often used stuff to local variables
local units = addon.Units
local specs = addon.Specs
local defaults = addon.Defaults
local CopyTable = addon.CopyTable
local LastEvent, LastEventTime, LastEventSource, LastEventSpell

local type = type
local pairs = pairs
local bband = bit.band
local smatch = string.match
local select = select
local wipe = wipe

local UnitGUID, UnitName, UnitIsEnemy, UnitIsPlayer = UnitGUID, UnitName, UnitIsEnemy, UnitIsPlayer
local UnitClass = UnitClass
local GetPlayerInfoByGUID = GetPlayerInfoByGUID
local GetTime = GetTime
local GetSpellInfo = GetSpellInfo
local RaidNotice_AddMessage, PlaySoundFile = RaidNotice_AddMessage, PlaySoundFile
local SendChatMessage = SendChatMessage

-- events checked
local events = {
--	["PLAYER_ENTERING_WORLD"] = true,
	["ZONE_CHANGED_NEW_AREA"] = true,
 	["COMBAT_LOG_EVENT_UNFILTERED"] = true,
	["PLAYER_TARGET_CHANGED"] = true,
	["PLAYER_FOCUS_CHANGED"] = true,
	["ARENA_OPPONENT_UPDATE"] = true,
}

-- combatlog event subtypes checked
local eventtypes = {
	["SPELL_CAST_SUCCESS"] = true,
}

local GUID_INVALID = "0x0000000000000000"

-- saves manual changes to anchors
local function OnAnchorChange(event, anchor, ...)
	local name = addon:GetAnchorKey(anchor.name)
	if event == "FancyBar_GroupResize" then
		addon.DB.profile.Anchors[name].Width = ...
		if name == "Anchor_Arena" then
			addon:UpdateAnchorStyle(name)
		end
	elseif event == "FancyBar_GroupMove" then
		local p, f, r, x, y = ...
		if name == "Anchor_Arena" then
			local dx = addon.DB.profile.Anchors[name].Position.NextX
			local dy = addon.DB.profile.Anchors[name].Position.NextY

			if anchor.name == "Anchor_Arena2" then
				x = x - dx
				y = y - dy
			elseif anchor.name == "Anchor_Arena3" then
				x = x - dx*2
				y = y - dy*2
			elseif anchor.name == "Anchor_Arena4" then
				x = x - dx*3
				y = y - dy*3
			elseif anchor.name == "Anchor_Arena5" then
				x = x - dx*4
				y = y - dy*4
			end
		end
		addon.DB.profile.Anchors[name].Position.AttachPoint = p
		addon.DB.profile.Anchors[name].Position.AttachFrame = f
		addon.DB.profile.Anchors[name].Position.AttachRelative = r
		addon.DB.profile.Anchors[name].Position.AttachX = x
		addon.DB.profile.Anchors[name].Position.AttachY = y
		if name == "Anchor_Arena" then
			addon:UpdateAnchorStyle(name)
		end
	end
end

-- converts an RGBA color to a table
local function ConvertColor(color)
	if color and color.r then
		return {color.r, color.g, color.b, color.a}
	else
		return nil
	end
end

-- restyles bars
local function StyleBar(bar)
	local anchor = bar.group
	if not anchor then return end
	local anchorname = addon:GetAnchorKey(anchor.name)

	local db = addon.DB.profile.Anchors[anchorname]

	if media:IsValid("statusbar", db.Bars.Texture) then
		bar:SetTexture(media:Fetch("statusbar", db.Bars.Texture))
	end
	bar:SetLabelFont(media:Fetch("font", db.Bars.NameFont), db.Bars.NameFontSize)
	bar:SetDurationFont(media:Fetch("font", db.Bars.TimeFont), db.Bars.TimeFontSize)
	bar:SetGhost(db.Bars.Ghost, nil, db.Bars.GhostMessage, ConvertColor(db.Bars.GhostBackground))
	bar:SetFadeout(db.Bars.Fadeout, nil, db.Bars.FadeoutMessage, ConvertColor(db.Bars.FadeoutBackground))
end

function addon:OnInitialize()
	--register chat commands
	addon:RegisterChatCommand("pvptimer", "ChatCommand")
	addon:RegisterChatCommand("pt", "ChatCommand")
	
	--create db
	addon.DB = LibStub("AceDB-3.0"):New("PTDB", defaults, true)
	local db = addon.DB

	--create options loader
	ACR:RegisterOptionsTable("PvPTimer", addon.Options)
	addon.OptionsFrame = ACD:AddToBlizOptions("PvPTimer", "PvPTimer", nil, "Settings")
	addon.Profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(db)

	addon.DB.RegisterCallback(addon, "OnProfileChanged", "OnEnable")
	addon.DB.RegisterCallback(addon, "OnProfileCopied", "OnEnable")
	addon.DB.RegisterCallback(addon, "OnProfileReset", "OnEnable")

	ACR:RegisterOptionsTable("PvPTimer-Profiles", addon.Profiles)
	ACD:AddToBlizOptions("PvPTimer-Profiles", "Profiles", "PvPTimer")
	-- local LibDualSpec = LibStub("LibDualSpec-1.0")
	-- LibDualSpec:EnhanceOptions(addon.Profiles, addon.DB)

	-- create default anchors
	local anchors = addon.Anchors
	anchors["Anchor_Target"] = fb:NewGroup(300, "Target", "Anchor_Target")
	anchors["Anchor_Focus"] = fb:NewGroup(300, "Focus", "Anchor_Focus")
	anchors["Anchor_Arena1"] = fb:NewGroup(300, "Arena #1", "Anchor_Arena1")
	anchors["Anchor_Arena2"] = fb:NewGroup(300, "Arena #2", "Anchor_Arena2")
	anchors["Anchor_Arena3"] = fb:NewGroup(300, "Arena #3", "Anchor_Arena3")
	anchors["Anchor_Arena4"] = fb:NewGroup(300, "Arena #4", "Anchor_Arena4")
	anchors["Anchor_Arena5"] = fb:NewGroup(300, "Arena #5", "Anchor_Arena5")

	-- temporary hack for drag handles
	for k, v in pairs(anchors) do
		v.FG_DragHandle:SetTexture("Interface\\Addons\PvPTimer\\Textures\\draghandle")
	end

	-- create spec icons
	local icons = addon.AnchorIcons
	icons["Anchor_Target"] = anchors["Anchor_Target"]:CreateTexture(nil, "OVERLAY")
	icons["Anchor_Focus"] = anchors["Anchor_Target"]:CreateTexture(nil, "OVERLAY")
	icons["Anchor_Arena1"] = anchors["Anchor_Target"]:CreateTexture(nil, "OVERLAY")
	icons["Anchor_Arena2"] = anchors["Anchor_Target"]:CreateTexture(nil, "OVERLAY")
	icons["Anchor_Arena3"] = anchors["Anchor_Target"]:CreateTexture(nil, "OVERLAY")
	icons["Anchor_Arena4"] = anchors["Anchor_Target"]:CreateTexture(nil, "OVERLAY")
	icons["Anchor_Arena5"] = anchors["Anchor_Target"]:CreateTexture(nil, "OVERLAY")

	for k, v in pairs(anchors) do
		v.RegisterCallback(v, "FancyBar_GroupResize", OnAnchorChange)
		v.RegisterCallback(v, "FancyBar_GroupMove", OnAnchorChange)
	end
	
	-- all done
	addon:Print(L["Addon loaded."])
end

function addon:OnEnable()
	local db = addon.DB
	local anchors = addon.Anchors

	--copy any missing options from default settings
	for a in pairs(anchors) do
		a = addon:GetAnchorKey(a)
		for k, v in pairs(db.global.anchor_default) do
			if db.profile.Anchors[a][k] == nil or db.profile.Anchors[a][k] == {} then
				db.profile.Anchors[a][k] = addon:CopyTable(v)
			end
		end
		for k, v in pairs(db.global.anchor_default.Bars) do
			if db.profile.Anchors[a].Bars[k] == nil or db.profile.Anchors[a].Bars[k] == {} then
				db.profile.Anchors[a].Bars[k] = addon:CopyTable(v)
			end
		end
		for k, v in pairs(db.global.anchor_default.Position) do
			if db.profile.Anchors[a].Position[k] == nil or db.profile.Anchors[a].Position[k] == {} then
				db.profile.Anchors[a].Position[k] = addon:CopyTable(v)
			end
		end
	end

	-- register events
	addon:UnregisterAllEvents()
	for k in pairs(events) do
		addon:RegisterEvent(k)
	end

	-- update anchors
	addon:RefreshAnchors()
end

function addon:ChatCommand(input)
	if input == "sc" or input == "spell" or input == "spellconfig" then
		-- open spell configuration
		addon:ToggleSpellConfig()
	elseif input == "lock" or input == "unlock" then
		-- open spell configuration
		addon:ToggleLock()
	else
		-- open blizzard config
		InterfaceOptionsFrame_OpenToCategory(addon.OptionsFrame)
	end
end

function addon:ToggleLock()
	addon.AnchorsLocked = not addon.AnchorsLocked
	addon:RefreshAnchors()
end

function addon:RunTest()
	for k, v in pairs(addon.Anchors) do
		v:RunTest(5)
	end
	addon:RefreshAnchors()
end

-- Refresh all anchors
function addon:RefreshAnchors(noStyle)
	local db = addon.DB.profile.Anchors
	local anchors = addon.Anchors

	-- move, size, style anchors
	if not noStyle then
		addon:UpdateAnchorStyle("Anchor_Target")
		addon:UpdateAnchorStyle("Anchor_Focus")
		addon:UpdateAnchorStyle("Anchor_Arena")
	end

	addon:PLAYER_TARGET_CHANGED()
	addon:PLAYER_FOCUS_CHANGED()
	addon:ARENA_OPPONENT_UPDATE()
end

function addon:UpdateAnchorStyle(anchorname)
	local db = addon.DB.profile.Anchors
	local anchors = {}

	if anchorname == "Anchor_Arena" then
		for i = 1, 5 do
			anchors["Anchor_Arena"..i] = db[anchorname]
		end
	else
		anchors = {[anchorname] = db[anchorname]}
	end

	for k, a in pairs(anchors) do
		local v = addon.Anchors[k]
		local ap = a.Position
		local ab = a.Bars

		if addon:GetZoneSetting("Anchors", anchorname) == false then
			v:Clear(true)
			v:Lock(false)
			v:Hide()
		else
			v:Show()
			v:Lock(addon.AnchorsLocked)

			v:ClearAllPoints()
			v:SetPoint(ap.AttachPoint, ap.AttachFrame, ap.AttachRelative, ap.AttachX, ap.AttachY)
			v:SetWidth(a.Width)
			v:SetScale(a.Scale)
			v:SetColumns(ab.Columns)
			v:SetVisibleBars(ab.MaxBars)
			v:SetStartPoint(ab.StartPoint)
			if ab.IconSize < 8 then ab.IconSize = 8 end
			if ab.IconMode then
				v:SetIconMode(ab.IconSize)
			else
				v:SetIconMode(false)
			end
		end
		addon:UpdateAnchor(v)
	end
	if anchorname == "Anchor_Arena" then
		local x = db["Anchor_Arena"].Position.NextX
		local y = db["Anchor_Arena"].Position.NextY
		addon.Anchors["Anchor_Arena2"]:ClearAllPoints()
		addon.Anchors["Anchor_Arena3"]:ClearAllPoints()
		addon.Anchors["Anchor_Arena4"]:ClearAllPoints()
		addon.Anchors["Anchor_Arena5"]:ClearAllPoints()
		addon.Anchors["Anchor_Arena2"]:SetPoint("CENTER", addon.Anchors["Anchor_Arena1"], "CENTER", x, y)
		addon.Anchors["Anchor_Arena3"]:SetPoint("CENTER", addon.Anchors["Anchor_Arena2"], "CENTER", x, y)
		addon.Anchors["Anchor_Arena4"]:SetPoint("CENTER", addon.Anchors["Anchor_Arena3"], "CENTER", x, y)
		addon.Anchors["Anchor_Arena5"]:SetPoint("CENTER", addon.Anchors["Anchor_Arena4"], "CENTER", x, y)
	end
end

-- updates all spells on an anchor
function addon:UpdateAnchor(anchor, spells, GUID)
	local anchorname = addon:GetAnchorKey(anchor.name)
	local db = addon.DB.profile.Anchors[anchorname]

	-- redraw spell timer bars
	for k in pairs(anchor.bars) do
		if k:GetData("id") then
			k:Stop(true)
		else
			StyleBar(k)
		end
	end

	local unit = UnitGUID(addon.Const.Units[anchor.name])
	local class = select(2, UnitClass(addon.Const.Units[anchor.name]))
	local spec
	if addon.Units[unit] then
		spec = addon:GetSpec(class, addon.Units[unit].spec)
	end
	local icon = addon.AnchorIcons[anchor.name]
	if spec and db.ShowSpec then
		icon:SetTexture(spec.icon)
		icon:SetWidth(db.ShowSpec)
		icon:SetHeight(db.ShowSpec)
		addon:SetIconPoint(icon, anchor, db.SpecPosition, db.Bars.StartPoint)
		icon:Show()
	else
		icon:Hide()
	end

	if not spells or not GUID then return end
	for k, v in pairs(spells) do
		if type(k) == "number" then
			local spell = addon:GetSpell(k)
			if not spell then break end
			local spec = spell.spec or 0
			local cd = addon:GetSpellCooldown(k, spec)
			local duration = v + cd - GetTime()
			if duration>0 then
				addon:UpdateSpell(anchor, k, duration, cd)
			end
		end
	end
end

-- sends alert messages
function addon:SendAlert(event, src, dst, spell, spellType)
	if not event or not src or not spell then return end
	local db = addon.DB.profile.Alert
	if not db[spellType] then spellType = 'misc' end

	-- check if alerts are enabled for this spell type
	if db[spellType].Enabled == false then return end

	local srcName = select(6, GetPlayerInfoByGUID(src))
	local dstName
	if dst and dst ~= GUID_INVALID then dstName = select(6, GetPlayerInfoByGUID(dst)) end
	local name
	if spell and type(spell) == "number" then name = GetSpellInfo(spell) end
	local zonetype = select(2, IsInInstance())
	local rated = IsRatedBattleground()
	local target, onlytarget = false, false
	
	-- check current zone and send alert accordingly
	if zonetype == "pvp" and not rated then -- we're in a normal bg
		target = db[spellType].BG
		onlytarget = db[spellType].BG_OnlyTarget
	elseif zonetype == "pvp" and rated then -- we're in a rated bg
		target = db[spellType].RatedBG
		onlytarget = db[spellType].RatedBG_OnlyTarget
	elseif zonetype == "arena" then -- we're in an arena
		target = db[spellType].Arena
		onlytarget = db[spellType].Arena_OnlyTarget
	else -- we're somewhere else
		target = db[spellType].Else
		onlytarget = db[spellType].Else_OnlyTarget
	end
	
	-- no info or alert target is 'none'
	if target == '_none_' then return end
	-- if 'Only target/focus' is on, check if target/focus guid is same as source
	if onlytarget and (UnitGUID("target") ~= src or UnitGUID("focus") ~= src) then return end
	
	local message
	if event == "used" then
		message = srcName.." "..L["USED"].." "..name
	elseif event == "spec" then
		message = srcName.." "..L["SPECCED"].." "..spell
	end

	if target == "screen" then
		RaidNotice_AddMessage(RaidWarningFrame, message, ChatTypeInfo["RAID_WARNING"])		
		PlaySoundFile("Sound\\Interface\\RaidWarning.wav")
	elseif target == "screen_nosound" then
		RaidNotice_AddMessage(RaidWarningFrame, message, ChatTypeInfo["RAID_WARNING"])		
	elseif target == "party" then
		SendChatMessage(message, "PARTY")
	elseif target == "raid" then
		SendChatMessage(message, "RAID")
	elseif target == "rw" then
		SendChatMessage(message, "RAID_WARNING")
	elseif target == "bg" then
		SendChatMessage(message, "BATTLEGROUND")
	else
		addon:Print(message)
	end
end

-- sends spec detected messages
function addon:AlertSpec(source, spec)
	local _, class, _, _, _, name = GetPlayerInfoByGUID(source)
	local info = addon:GetSpec(class, spec)
	if info then
		addon:SendAlert("spec", source, nil, info.name, "spec")
	end
end

-- updates a spell timer bar
function addon:UpdateSpell(anchor, spellID, duration, maximum)
	-- exit if anchor is disabled
	local anchorname = addon:GetAnchorKey(anchor.name)
	if addon:GetZoneSetting("Anchors", anchorname) == false then return end

	local spell = addon:GetSpell(spellID)
	local color = addon:GetBarColor(anchorname, spell.type)

	-- check if bar exists, reuse it if it does
	local bar = anchor:Find("id", spellID)
	if not bar then
		bar = fb:NewBar(anchor, 64, 16)
		bar:SetData("id", spellID)
		bar:SetIcon(select(3, GetSpellInfo(spellID)))
		bar:SetDuration(duration, maximum)
		bar:Start()
	else
		bar:SetDuration(duration, maximum)
	end

	-- set label, scale and color
	bar:SetLabel(GetSpellInfo(spellID))
	bar:SetScale(anchor:GetScale())

	if color then
		bar:SetColor(ConvertColor(color))
	end
	-- apply any other styles
	StyleBar(bar)
end

--[[
function addon:PLAYER_ENTERING_WORLD()
	addon:ZONE_CHANGED_NEW_AREA()
end
]]

-- fired when entering a new area, clean up and redraw everything
function addon:ZONE_CHANGED_NEW_AREA()
	--wipe unit data
	wipe(addon.Units)
	
	addon:RefreshAnchors()
end

-- refresh target anchor when player target changes
function addon:PLAYER_TARGET_CHANGED()
	-- is anchor enabled?
	if addon:GetZoneSetting("Anchors", "Anchor_Target") == false then return end
	-- is it a valid enemy player?
	if UnitIsPlayer("target") and (debugmode or UnitIsEnemy("player", "target")) then
		local GUID = UnitGUID("target")
		local spells = units[GUID]
		addon:UpdateAnchor(addon.Anchors["Anchor_Target"], spells, GUID)
	else
		addon:UpdateAnchor(addon.Anchors["Anchor_Target"], nil, nil)
	end
end

-- refresh focus anchor when player focus changes
function addon:PLAYER_FOCUS_CHANGED()
	-- is anchor enabled?
	if addon:GetZoneSetting("Anchors", "Anchor_Focus") == false then return end
	-- is it a valid enemy player?
	if UnitIsPlayer("focus") and UnitIsEnemy("player", "focus") then
		local GUID = UnitGUID("focus")
		local spells = units[GUID]
		addon:UpdateAnchor(addon.Anchors["Anchor_Focus"], spells, GUID)
	else
		addon:UpdateAnchor(addon.Anchors["Anchor_Focus"], nil, nil)
	end
end

-- refresh arena anchors as opponent info updates
function addon:ARENA_OPPONENT_UPDATE()
	-- is anchor enabled?
	if addon:GetZoneSetting("Anchors", "Anchor_Arena") == false then return end
	for i = 1, 5 do
		local unit = "arena"..i
		local anchor = "Anchor_Arena"..i
		if UnitIsPlayer(unit) then
			local GUID = UnitGUID(unit)
			local spells = units[GUID]
			addon:UpdateAnchor(addon.Anchors[anchor], spells, GUID)
		else
			addon:UpdateAnchor(addon.Anchors[anchor], nil, nil)
		end
	end
end

-- respond to combatlog events
function addon:COMBAT_LOG_EVENT_UNFILTERED(event, timeStamp, eventType, hideCaster, srcGUID, srcName, srcFlags, srcRaidFlags, dstGUID, dstName, dstFlags, dstRaidFlags, ...)
	-- check event subtype
	if not eventtypes[eventType] then return end

	local spellID, spellName, spellSchool, auraType = ...

	-- workaround for combat log bug (2 exact same entries generated for 1 event)
	if LastEvent == event and LastEventTime == timeStamp and LastEventSource == srcGUID and LastEventSpell == spellName then return end
	LastEvent = event
	LastEventTime = timeStamp
	LastEventSource = srcGUID
	LastEventSpell = spellName

	-- check if guid is valid and if unit is player and hostile
	if srcGUID == nil or srcGUID == GUID_INVALID or not GetPlayerInfoByGUID(srcGUID) or not (debugmode or addon:IsHostile(srcFlags)) then return end

	-- some additional failchecks
	if spellName == "-1" or spellID == 0 then return end

	local db = addon.DB.profile
	if not units[srcGUID] then
		units[srcGUID] = {}
	end
	
	-- spec detection #1 (spammable spells)
	if addon.SpecSpells[spellID] and not units[srcGUID].spec then
		units[srcGUID].spec = addon.SpecSpells[spellID]
		addon:AlertSpec(srcGUID, units[srcGUID].spec)
		addon:RefreshAnchors(true)
	end
	
	-- check if spell is in database
	local spell = addon:GetSpell(spellID)
	if not spell or not spell.cooldown or spell.cooldown<1 then return end

	local anchors = addon.Anchors
	local spec = spell.spec or false
	local cd = addon:GetSpellCooldown(spellID, spec)
	
	-- spec detection #2 (cd spells)
	if spec and not units[srcGUID].spec then
		units[srcGUID].spec = spec
		addon:AlertSpec(srcGUID, spec)
		addon:RefreshAnchors(true)
	end
	
	if eventType == "SPELL_CAST_SUCCESS" then
		units[srcGUID][spellID] = GetTime()

		if srcGUID == UnitGUID("target") then
			addon:UpdateSpell(anchors.Anchor_Target, spellID, cd, cd)
		end

		if srcGUID == UnitGUID("focus") then
			addon:UpdateSpell(anchors.Anchor_Focus, spellID, cd, cd)
		end
		
		if not db.Anchors.Anchor_Arena.Enabled == false then
			for i = 1, 5 do
				local a = "Anchor_Arena"..i
				local u = "arena"..i
				if srcGUID == UnitGUID(u) then
					addon:UpdateSpell(anchors[a], spellID, cd, cd)
				end		
			end
		end

		addon:SendAlert("used", srcGUID, dstGUID, spellID, spell.type)
	end
end